@@ -1,8 +1,7 @@
 /* SPDX-License-Identifier: LGPL-2.1+ */
 
-#ifndef _GNU_SOURCE
-#define _GNU_SOURCE 1
-#endif
+#include "config.h"
+
 #include <caps.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -22,7 +21,6 @@
 #include "commands.h"
 #include "commands_utils.h"
 #include "conf.h"
-#include "config.h"
 #include "confile.h"
 #include "log.h"
 #include "lxc.h"
@@ -91,6 +89,7 @@ static const char *lxc_cmd_str(lxc_cmd_t cmd)
 		[LXC_CMD_GET_CGROUP_CTX]		= "get_cgroup_ctx",
 		[LXC_CMD_GET_CGROUP_FD]			= "get_cgroup_fd",
 		[LXC_CMD_GET_LIMIT_CGROUP_FD]		= "get_limit_cgroup_fd",
+		[LXC_CMD_GET_SYSTEMD_SCOPE]		= "get_systemd_scope",
 	};
 
 	if (cmd >= LXC_CMD_MAX)
@@ -126,7 +125,7 @@ static ssize_t lxc_cmd_rsp_recv_fds(int fd_sock, struct unix_fds *fds,
 
 	ret = lxc_abstract_unix_recv_fds(fd_sock, fds, rsp, sizeof(*rsp));
 	if (ret < 0)
-		return log_error(ret, "Failed to receive file descriptors");
+		return log_error(ret, "Failed to receive file descriptors for command \"%s\"", cur_cmdstr);
 
 	/*
 	 * If we end up here with fewer or more file descriptors the caller
@@ -135,16 +134,16 @@ static ssize_t lxc_cmd_rsp_recv_fds(int fd_sock, struct unix_fds *fds,
 	 */
 
 	if (fds->flags & UNIX_FDS_RECEIVED_EXACT)
-		return log_debug(ret, "Received exact number of file descriptors %u == %u",
-				 fds->fd_count_max, fds->fd_count_ret);
+		return log_debug(ret, "Received exact number of file descriptors %u == %u for command \"%s\"",
+				 fds->fd_count_max, fds->fd_count_ret, cur_cmdstr);
 
 	if (fds->flags & UNIX_FDS_RECEIVED_LESS)
-		return log_debug(ret, "Received less file descriptors %u < %u",
-				 fds->fd_count_ret, fds->fd_count_max);
+		return log_debug(ret, "Received less file descriptors %u < %u for command \"%s\"",
+				 fds->fd_count_ret, fds->fd_count_max, cur_cmdstr);
 
 	if (fds->flags & UNIX_FDS_RECEIVED_MORE)
-		return log_debug(ret, "Received more file descriptors (excessive fds were automatically closed) %u > %u",
-				 fds->fd_count_ret, fds->fd_count_max);
+		return log_debug(ret, "Received more file descriptors (excessive fds were automatically closed) %u > %u for command \"%s\"",
+				 fds->fd_count_ret, fds->fd_count_max, cur_cmdstr);
 
 	DEBUG("Command \"%s\" received response", cur_cmdstr);
 	return ret;
@@ -213,7 +212,11 @@ static ssize_t lxc_cmd_rsp_recv(int sock, struct lxc_cmd_rr *cmd)
 		break;
 	case LXC_CMD_GET_CGROUP_CTX:
 		fds->fd_count_max = CGROUP_CTX_MAX_FD;
-		fds->flags |= UNIX_FDS_ACCEPT_LESS;
+		/* 
+		 * The container might run without any cgroup support at all,
+		 * i.e. no writable cgroup hierarchy was found.
+		 */
+		fds->flags |= UNIX_FDS_ACCEPT_LESS  | UNIX_FDS_ACCEPT_NONE ;
 		break;
 	default:
 		fds->fd_count_max = 0;
@@ -253,14 +256,14 @@ static ssize_t lxc_cmd_rsp_recv(int sock, struct lxc_cmd_rr *cmd)
 	case LXC_CMD_GET_CGROUP_FD:		/* data */
 		__fallthrough;
 	case LXC_CMD_GET_LIMIT_CGROUP_FD:	/* data */
-		if (rsp->datalen > sizeof(struct cgroup_fd))
+		if ((size_t)rsp->datalen > sizeof(struct cgroup_fd))
 			return syserror_set(-EINVAL, "Invalid response size from server for \"%s\"", cur_cmdstr);
 
 		/* Don't pointlessly allocate. */
 		rsp->data = (void *)cmd->req.data;
 		break;
 	case LXC_CMD_GET_CGROUP_CTX:		/* data */
-		if (rsp->datalen > sizeof(struct cgroup_ctx))
+		if ((size_t)rsp->datalen > sizeof(struct cgroup_ctx))
 			return syserror_set(-EINVAL, "Invalid response size from server for \"%s\"", cur_cmdstr);
 
 		/* Don't pointlessly allocate. */
@@ -611,7 +614,7 @@ pid_t lxc_cmd_get_init_pid(const char *name, const char *lxcpath)
 
 static int lxc_cmd_get_init_pid_callback(int fd, struct lxc_cmd_req *req,
 					 struct lxc_handler *handler,
-					 struct lxc_epoll_descr *descr)
+					 struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.data = PID_TO_PTR(handler->pid),
@@ -648,7 +651,7 @@ int lxc_cmd_get_init_pidfd(const char *name, const char *lxcpath)
 
 static int lxc_cmd_get_init_pidfd_callback(int fd, struct lxc_cmd_req *req,
 					   struct lxc_handler *handler,
-					   struct lxc_epoll_descr *descr)
+					   struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.ret = -EBADF,
@@ -688,7 +691,7 @@ int lxc_cmd_get_devpts_fd(const char *name, const char *lxcpath)
 
 static int lxc_cmd_get_devpts_fd_callback(int fd, struct lxc_cmd_req *req,
 					  struct lxc_handler *handler,
-					  struct lxc_epoll_descr *descr)
+					  struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.ret = -EBADF,
@@ -703,7 +706,7 @@ static int lxc_cmd_get_devpts_fd_callback(int fd, struct lxc_cmd_req *req,
 
 int lxc_cmd_get_seccomp_notify_fd(const char *name, const char *lxcpath)
 {
-#ifdef HAVE_SECCOMP_NOTIFY
+#if HAVE_DECL_SECCOMP_NOTIFY_FD
 	bool stopped = false;
 	int fd;
 	ssize_t ret;
@@ -732,9 +735,9 @@ int lxc_cmd_get_seccomp_notify_fd(const char *name, const char *lxcpath)
 
 static int lxc_cmd_get_seccomp_notify_fd_callback(int fd, struct lxc_cmd_req *req,
 						  struct lxc_handler *handler,
-						  struct lxc_epoll_descr *descr)
+						  struct lxc_async_descr *descr)
 {
-#ifdef HAVE_SECCOMP_NOTIFY
+#if HAVE_DECL_SECCOMP_NOTIFY_FD
 	struct lxc_cmd_rsp rsp = {
 		.ret = -EBADF,
 	};
@@ -764,16 +767,21 @@ int lxc_cmd_get_cgroup_ctx(const char *name, const char *lxcpath,
 		return sysdebug("Failed to process \"%s\"",
 				lxc_cmd_str(LXC_CMD_GET_CGROUP_CTX));
 
-	if (cmd.rsp.ret < 0)
+	if (cmd.rsp.ret < 0) {
+		/* Container does not have any writable cgroups. */
+		if (ret_ctx->fd_len == 0)
+			return 0;
+
 		return sysdebug_set(cmd.rsp.ret, "Failed to receive file descriptor for \"%s\"",
 				    lxc_cmd_str(LXC_CMD_GET_CGROUP_CTX));
+	}
 
 	return 0;
 }
 
 static int lxc_cmd_get_cgroup_ctx_callback(int fd, struct lxc_cmd_req *req,
 					   struct lxc_handler *handler,
-					   struct lxc_epoll_descr *descr)
+					   struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.ret = EINVAL,
@@ -824,7 +832,7 @@ int lxc_cmd_get_clone_flags(const char *name, const char *lxcpath)
 
 static int lxc_cmd_get_clone_flags_callback(int fd, struct lxc_cmd_req *req,
 					    struct lxc_handler *handler,
-					    struct lxc_epoll_descr *descr)
+					    struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.data = INT_TO_PTR(handler->ns_clone_flags),
@@ -914,7 +922,7 @@ char *lxc_cmd_get_limit_cgroup_path(const char *name, const char *lxcpath,
 
 static int __lxc_cmd_get_cgroup_callback(int fd, struct lxc_cmd_req *req,
 					 struct lxc_handler *handler,
-					 struct lxc_epoll_descr *descr,
+					 struct lxc_async_descr *descr,
 					 bool limiting_cgroup)
 {
 	ssize_t ret;
@@ -950,14 +958,14 @@ static int __lxc_cmd_get_cgroup_callback(int fd, struct lxc_cmd_req *req,
 
 static int lxc_cmd_get_cgroup_callback(int fd, struct lxc_cmd_req *req,
 				       struct lxc_handler *handler,
-				       struct lxc_epoll_descr *descr)
+				       struct lxc_async_descr *descr)
 {
 	return __lxc_cmd_get_cgroup_callback(fd, req, handler, descr, false);
 }
 
 static int lxc_cmd_get_limit_cgroup_callback(int fd, struct lxc_cmd_req *req,
 					     struct lxc_handler *handler,
-					     struct lxc_epoll_descr *descr)
+					     struct lxc_async_descr *descr)
 {
 	return __lxc_cmd_get_cgroup_callback(fd, req, handler, descr, true);
 }
@@ -997,7 +1005,7 @@ char *lxc_cmd_get_config_item(const char *name, const char *item,
 
 static int lxc_cmd_get_config_item_callback(int fd, struct lxc_cmd_req *req,
 					    struct lxc_handler *handler,
-					    struct lxc_epoll_descr *descr)
+					    struct lxc_async_descr *descr)
 {
 	__do_free char *cidata = NULL;
 	int cilen;
@@ -1059,7 +1067,7 @@ int lxc_cmd_get_state(const char *name, const char *lxcpath)
 
 static int lxc_cmd_get_state_callback(int fd, struct lxc_cmd_req *req,
 				      struct lxc_handler *handler,
-				      struct lxc_epoll_descr *descr)
+				      struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.data = INT_TO_PTR(handler->state),
@@ -1104,7 +1112,7 @@ int lxc_cmd_stop(const char *name, const char *lxcpath)
 
 static int lxc_cmd_stop_callback(int fd, struct lxc_cmd_req *req,
 				 struct lxc_handler *handler,
-				 struct lxc_epoll_descr *descr)
+				 struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp;
 	int stopsignal = SIGKILL;
@@ -1155,7 +1163,7 @@ int lxc_cmd_terminal_winch(const char *name, const char *lxcpath)
 
 static int lxc_cmd_terminal_winch_callback(int fd, struct lxc_cmd_req *req,
 					   struct lxc_handler *handler,
-					   struct lxc_epoll_descr *descr)
+					   struct lxc_async_descr *descr)
 {
 	/* should never be called */
 	return syserror_set(-ENOSYS, "Called lxc_cmd_terminal_winch_callback()");
@@ -1207,7 +1215,7 @@ int lxc_cmd_get_tty_fd(const char *name, int *ttynum, int *fd, const char *lxcpa
 
 static int lxc_cmd_get_tty_fd_callback(int fd, struct lxc_cmd_req *req,
 				       struct lxc_handler *handler,
-				       struct lxc_epoll_descr *descr)
+				       struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.ret = -EBADF,
@@ -1258,7 +1266,7 @@ char *lxc_cmd_get_name(const char *hashed_sock_name)
 
 static int lxc_cmd_get_name_callback(int fd, struct lxc_cmd_req *req,
 				     struct lxc_handler *handler,
-				     struct lxc_epoll_descr *descr)
+				     struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp;
 
@@ -1298,7 +1306,7 @@ char *lxc_cmd_get_lxcpath(const char *hashed_sock_name)
 
 static int lxc_cmd_get_lxcpath_callback(int fd, struct lxc_cmd_req *req,
 					struct lxc_handler *handler,
-					struct lxc_epoll_descr *descr)
+					struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.ret		= 0,
@@ -1309,6 +1317,55 @@ static int lxc_cmd_get_lxcpath_callback(int fd, struct lxc_cmd_req *req,
 	return lxc_cmd_rsp_send_reap(fd, &rsp);
 }
 
+char *lxc_cmd_get_systemd_scope(const char *name, const char *lxcpath)
+{
+	bool stopped = false;
+	ssize_t ret;
+	struct lxc_cmd_rr cmd;
+
+	lxc_cmd_init(&cmd, LXC_CMD_GET_SYSTEMD_SCOPE);
+
+	ret = lxc_cmd(name, &cmd, &stopped, lxcpath, NULL);
+	if (ret < 0)
+		return NULL;
+
+	if (cmd.rsp.ret == 0)
+		return cmd.rsp.data;
+
+	return NULL;
+}
+
+static int lxc_cmd_get_systemd_scope_callback(int fd, struct lxc_cmd_req *req,
+					     struct lxc_handler *handler,
+					     struct lxc_async_descr *descr)
+{
+	__do_free char *scope = NULL;
+	struct lxc_cmd_rsp rsp = {
+		.ret = -EINVAL,
+	};
+
+	// cgroup_meta.systemd_scope is the full cgroup path to the scope.
+	// The caller just wants the actual scope name, that is, basename().
+	// (XXX - or do we want the caller to massage it?  I'm undecided)
+	if (handler->conf->cgroup_meta.systemd_scope) {
+		scope = strrchr(handler->conf->cgroup_meta.systemd_scope, '/');
+		if (scope && *scope)
+			scope++;
+		if (scope && *scope)
+			scope = strdup(scope);
+	}
+
+	if (!scope)
+		goto out;
+
+	rsp.ret = 0;
+	rsp.data = scope;
+	rsp.datalen = strlen(scope) + 1;
+
+out:
+	return lxc_cmd_rsp_send_reap(fd, &rsp);
+}
+
 int lxc_cmd_add_state_client(const char *name, const char *lxcpath,
 			     lxc_state_t states[static MAX_STATE],
 			     int *state_client_fd)
@@ -1351,7 +1408,7 @@ int lxc_cmd_add_state_client(const char *name, const char *lxcpath,
 
 static int lxc_cmd_add_state_client_callback(__owns int fd, struct lxc_cmd_req *req,
 					     struct lxc_handler *handler,
-					     struct lxc_epoll_descr *descr)
+					     struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.ret = -EINVAL,
@@ -1403,7 +1460,7 @@ int lxc_cmd_add_bpf_device_cgroup(const char *name, const char *lxcpath,
 
 static int lxc_cmd_add_bpf_device_cgroup_callback(int fd, struct lxc_cmd_req *req,
 						  struct lxc_handler *handler,
-						  struct lxc_epoll_descr *descr)
+						  struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {
 		.ret = -EINVAL,
@@ -1470,7 +1527,7 @@ int lxc_cmd_console_log(const char *name, const char *lxcpath,
 
 static int lxc_cmd_console_log_callback(int fd, struct lxc_cmd_req *req,
 					struct lxc_handler *handler,
-					struct lxc_epoll_descr *descr)
+					struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp;
 	uint64_t buffer_size = handler->conf->console.buffer_size;
@@ -1489,7 +1546,7 @@ static int lxc_cmd_console_log_callback(int fd, struct lxc_cmd_req *req,
 	if (log->read)
 		rsp.data = lxc_ringbuf_get_read_addr(buf);
 
-	if (log->read_max > 0 && (log->read_max <= rsp.datalen))
+	if (log->read_max > 0 && (log->read_max <= (uint64_t)rsp.datalen))
 		rsp.datalen = log->read_max;
 
 	/* there's nothing to read */
@@ -1526,7 +1583,7 @@ int lxc_cmd_serve_state_clients(const char *name, const char *lxcpath,
 
 static int lxc_cmd_serve_state_clients_callback(int fd, struct lxc_cmd_req *req,
 						struct lxc_handler *handler,
-						struct lxc_epoll_descr *descr)
+						struct lxc_async_descr *descr)
 {
 	int ret;
 	lxc_state_t state = PTR_TO_INT(req->data);
@@ -1545,7 +1602,7 @@ int lxc_cmd_seccomp_notify_add_listener(const char *name, const char *lxcpath,
 					/* unused */ unsigned int flags)
 {
 
-#ifdef HAVE_SECCOMP_NOTIFY
+#if HAVE_DECL_SECCOMP_NOTIFY_FD
 	bool stopped = false;
 	ssize_t ret;
 	struct lxc_cmd_rr cmd;
@@ -1566,11 +1623,11 @@ int lxc_cmd_seccomp_notify_add_listener(const char *name, const char *lxcpath,
 static int lxc_cmd_seccomp_notify_add_listener_callback(int fd,
 							struct lxc_cmd_req *req,
 							struct lxc_handler *handler,
-							struct lxc_epoll_descr *descr)
+							struct lxc_async_descr *descr)
 {
 	struct lxc_cmd_rsp rsp = {0};
 
-#ifdef HAVE_SECCOMP_NOTIFY
+#if HAVE_DECL_SECCOMP_NOTIFY_FD
 	int ret;
 	__do_close int recv_fd = -EBADF;
 
@@ -1587,8 +1644,10 @@ static int lxc_cmd_seccomp_notify_add_listener_callback(int fd,
 		goto out;
 	}
 
-	ret = lxc_mainloop_add_handler(descr, recv_fd, seccomp_notify_handler,
-				       handler);
+	ret = lxc_mainloop_add_handler(descr, recv_fd,
+				       seccomp_notify_handler,
+				       seccomp_notify_cleanup_handler,
+				       handler, "seccomp_notify_handler");
 	if (ret < 0) {
 		rsp.ret = -errno;
 		goto out;
@@ -1621,7 +1680,7 @@ int lxc_cmd_freeze(const char *name, const char *lxcpath, int timeout)
 
 static int lxc_cmd_freeze_callback(int fd, struct lxc_cmd_req *req,
 				   struct lxc_handler *handler,
-				   struct lxc_epoll_descr *descr)
+				   struct lxc_async_descr *descr)
 {
 	int timeout = PTR_TO_INT(req->data);
 	struct lxc_cmd_rsp rsp = {
@@ -1653,7 +1712,7 @@ int lxc_cmd_unfreeze(const char *name, const char *lxcpath, int timeout)
 
 static int lxc_cmd_unfreeze_callback(int fd, struct lxc_cmd_req *req,
 				   struct lxc_handler *handler,
-				   struct lxc_epoll_descr *descr)
+				   struct lxc_async_descr *descr)
 {
 	int timeout = PTR_TO_INT(req->data);
 	struct lxc_cmd_rsp rsp = {
@@ -1713,7 +1772,7 @@ int lxc_cmd_get_limit_cgroup_fd(const char *name, const char *lxcpath,
 
 static int __lxc_cmd_get_cgroup_fd_callback(int fd, struct lxc_cmd_req *req,
 					    struct lxc_handler *handler,
-					    struct lxc_epoll_descr *descr,
+					    struct lxc_async_descr *descr,
 					    bool limit)
 {
 	struct lxc_cmd_rsp rsp = {
@@ -1745,14 +1804,14 @@ static int __lxc_cmd_get_cgroup_fd_callback(int fd, struct lxc_cmd_req *req,
 
 static int lxc_cmd_get_cgroup_fd_callback(int fd, struct lxc_cmd_req *req,
 					  struct lxc_handler *handler,
-					  struct lxc_epoll_descr *descr)
+					  struct lxc_async_descr *descr)
 {
 	return __lxc_cmd_get_cgroup_fd_callback(fd, req, handler, descr, false);
 }
 
 static int lxc_cmd_get_limit_cgroup_fd_callback(int fd, struct lxc_cmd_req *req,
 						struct lxc_handler *handler,
-						struct lxc_epoll_descr *descr)
+						struct lxc_async_descr *descr)
 {
 	return __lxc_cmd_get_cgroup_fd_callback(fd, req, handler, descr, true);
 }
@@ -1809,7 +1868,7 @@ int lxc_cmd_get_limit_cgroup2_fd(const char *name, const char *lxcpath)
 
 static int __lxc_cmd_get_cgroup2_fd_callback(int fd, struct lxc_cmd_req *req,
 					     struct lxc_handler *handler,
-					     struct lxc_epoll_descr *descr,
+					     struct lxc_async_descr *descr,
 					     bool limiting_cgroup)
 {
 	struct lxc_cmd_rsp rsp = {
@@ -1835,14 +1894,14 @@ static int __lxc_cmd_get_cgroup2_fd_callback(int fd, struct lxc_cmd_req *req,
 
 static int lxc_cmd_get_cgroup2_fd_callback(int fd, struct lxc_cmd_req *req,
 					   struct lxc_handler *handler,
-					   struct lxc_epoll_descr *descr)
+					   struct lxc_async_descr *descr)
 {
 	return __lxc_cmd_get_cgroup2_fd_callback(fd, req, handler, descr, false);
 }
 
 static int lxc_cmd_get_limit_cgroup2_fd_callback(int fd, struct lxc_cmd_req *req,
 						 struct lxc_handler *handler,
-						 struct lxc_epoll_descr *descr)
+						 struct lxc_async_descr *descr)
 {
 	return __lxc_cmd_get_cgroup2_fd_callback(fd, req, handler, descr, true);
 }
@@ -1859,10 +1918,10 @@ static int lxc_cmd_rsp_send_enosys(int fd, int id)
 
 static int lxc_cmd_process(int fd, struct lxc_cmd_req *req,
 			   struct lxc_handler *handler,
-			   struct lxc_epoll_descr *descr)
+			   struct lxc_async_descr *descr)
 {
 	typedef int (*callback)(int, struct lxc_cmd_req *, struct lxc_handler *,
-				struct lxc_epoll_descr *);
+				struct lxc_async_descr *);
 
 	callback cb[LXC_CMD_MAX] = {
 		[LXC_CMD_GET_TTY_FD]			= lxc_cmd_get_tty_fd_callback,
@@ -1891,6 +1950,7 @@ static int lxc_cmd_process(int fd, struct lxc_cmd_req *req,
 		[LXC_CMD_GET_CGROUP_CTX]		= lxc_cmd_get_cgroup_ctx_callback,
 		[LXC_CMD_GET_CGROUP_FD]			= lxc_cmd_get_cgroup_fd_callback,
 		[LXC_CMD_GET_LIMIT_CGROUP_FD]		= lxc_cmd_get_limit_cgroup_fd_callback,
+		[LXC_CMD_GET_SYSTEMD_SCOPE]		= lxc_cmd_get_systemd_scope_callback,
 	};
 
 	if (req->cmd >= LXC_CMD_MAX)
@@ -1900,28 +1960,17 @@ static int lxc_cmd_process(int fd, struct lxc_cmd_req *req,
 }
 
 static void lxc_cmd_fd_cleanup(int fd, struct lxc_handler *handler,
-			       struct lxc_epoll_descr *descr, const lxc_cmd_t cmd)
+			       const lxc_cmd_t cmd)
 {
-	lxc_terminal_free(handler->conf, fd);
-	lxc_mainloop_del_handler(descr, fd);
-
 	if (cmd == LXC_CMD_ADD_STATE_CLIENT) {
-		struct lxc_list *cur, *next;
-
-		lxc_list_for_each_safe(cur, &handler->conf->state_clients, next) {
-			struct lxc_state_client *client = cur->elem;
+		struct lxc_state_client *client, *nclient;
 
+		list_for_each_entry_safe(client, nclient, &handler->conf->state_clients, head) {
 			if (client->clientfd != fd)
 				continue;
 
-			/*
-			 * Only kick client from list so it can't be found
-			 * anymore. The actual close happens, as for all other
-			 * file descriptors, below.
-			 */
-			lxc_list_del(cur);
-			free(cur->elem);
-			free(cur);
+			list_del(&client->head);
+			free(client);
 
 			/*
 			 * No need to walk the whole list. If we found the state
@@ -1937,15 +1986,29 @@ static void lxc_cmd_fd_cleanup(int fd, struct lxc_handler *handler,
 		 * was already reached by the time we were ready to add it. So
 		 * fallthrough and clean it up.
 		 */
-		TRACE("Closing state client fd %d for command \"%s\"", fd, lxc_cmd_str(cmd));
+		TRACE("Deleted state client fd %d for command \"%s\"", fd, lxc_cmd_str(cmd));
 	}
 
-	TRACE("Closing client fd %d for command \"%s\"", fd, lxc_cmd_str(cmd));
+	/*
+	 * We're not closing the client fd here. They will instead be notified
+	 * from the mainloop when it calls the cleanup handler. This will cause
+	 * a slight delay but is semantically cleaner then what we used to do.
+	 */
+}
+
+static int lxc_cmd_cleanup_handler(int fd, void *data)
+{
+	struct lxc_handler *handler = data;
+
+	lxc_terminal_free(handler->conf, fd);
 	close(fd);
+	TRACE("Closing client fd %d for \"%s\"", fd, __FUNCTION__);
+	return 0;
+
 }
 
 static int lxc_cmd_handler(int fd, uint32_t events, void *data,
-			   struct lxc_epoll_descr *descr)
+			   struct lxc_async_descr *descr)
 {
 	__do_free void *reqdata = NULL;
 	int ret;
@@ -1965,20 +2028,20 @@ static int lxc_cmd_handler(int fd, uint32_t events, void *data,
 			__lxc_cmd_rsp_send(fd, &rsp);
 		}
 
-		goto out_close;
+		goto out;
 	}
 
 	if (ret == 0)
-		goto out_close;
+		goto out;
 
 	if (ret != sizeof(req)) {
 		WARN("Failed to receive full command request. Ignoring request for \"%s\"", lxc_cmd_str(req.cmd));
-		goto out_close;
+		goto out;
 	}
 
 	if ((req.datalen > LXC_CMD_DATA_MAX) && (req.cmd != LXC_CMD_CONSOLE_LOG)) {
 		ERROR("Received command data length %d is too large for command \"%s\"", req.datalen, lxc_cmd_str(req.cmd));
-		goto out_close;
+		goto out;
 	}
 
 	if (req.datalen > 0) {
@@ -1986,7 +2049,7 @@ static int lxc_cmd_handler(int fd, uint32_t events, void *data,
 		ret = lxc_recv_nointr(fd, reqdata, req.datalen, 0);
 		if (ret != req.datalen) {
 			WARN("Failed to receive full command request. Ignoring request for \"%s\"", lxc_cmd_str(req.cmd));
-			goto out_close;
+			goto out;
 		}
 
 		req.data = reqdata;
@@ -1995,24 +2058,24 @@ static int lxc_cmd_handler(int fd, uint32_t events, void *data,
 	ret = lxc_cmd_process(fd, &req, handler, descr);
 	if (ret < 0) {
 		DEBUG("Failed to process command %s; cleaning up client fd %d", lxc_cmd_str(req.cmd), fd);
-		goto out_close;
-	} else if (ret == LXC_CMD_REAP_CLIENT_FD) {
-		TRACE("Processed command %s; cleaning up client fd %d", lxc_cmd_str(req.cmd), fd);
-		goto out_close;
-	} else {
-		TRACE("Processed command %s; keeping client fd %d", lxc_cmd_str(req.cmd), fd);
+		goto out;
 	}
 
-out:
+	if (ret == LXC_CMD_REAP_CLIENT_FD) {
+		TRACE("Processed command %s; cleaning up client fd %d", lxc_cmd_str(req.cmd), fd);
+		goto out;
+	}
+
+	TRACE("Processed command %s; keeping client fd %d", lxc_cmd_str(req.cmd), fd);
 	return LXC_MAINLOOP_CONTINUE;
 
-out_close:
-	lxc_cmd_fd_cleanup(fd, handler, descr, req.cmd);
-	goto out;
+out:
+	lxc_cmd_fd_cleanup(fd, handler, req.cmd);
+	return LXC_MAINLOOP_DISARM;
 }
 
 static int lxc_cmd_accept(int fd, uint32_t events, void *data,
-			  struct lxc_epoll_descr *descr)
+			  struct lxc_async_descr *descr)
 {
 	__do_close int connection = -EBADF;
 	int opt = 1, ret = -1;
@@ -2029,7 +2092,10 @@ static int lxc_cmd_accept(int fd, uint32_t events, void *data,
 	if (ret < 0)
 		return log_error_errno(ret, errno, "Failed to enable necessary credentials on command socket");
 
-	ret = lxc_mainloop_add_handler(descr, connection, lxc_cmd_handler, data);
+	ret = lxc_mainloop_add_oneshot_handler(descr, connection,
+					       lxc_cmd_handler,
+					       lxc_cmd_cleanup_handler,
+					       data, "lxc_cmd_handler");
 	if (ret)
 		return log_error(ret, "Failed to add command handler");
 
@@ -2063,12 +2129,15 @@ int lxc_server_init(const char *name, const char *lxcpath, const char *suffix)
 	return log_trace(move_fd(fd), "Created abstract unix socket \"%s\"", &path[1]);
 }
 
-int lxc_cmd_mainloop_add(const char *name, struct lxc_epoll_descr *descr,
+int lxc_cmd_mainloop_add(const char *name, struct lxc_async_descr *descr,
 			 struct lxc_handler *handler)
 {
 	int ret;
 
-	ret = lxc_mainloop_add_handler(descr, handler->conf->maincmd_fd, lxc_cmd_accept, handler);
+	ret = lxc_mainloop_add_handler(descr, handler->conf->maincmd_fd,
+				       lxc_cmd_accept,
+				       default_cleanup_handler,
+				       handler, "lxc_cmd_accept");
 	if (ret < 0)
 		return log_error(ret, "Failed to add handler for command socket fd %d", handler->conf->maincmd_fd);
 
