@@ -1,9 +1,7 @@
 /* SPDX-License-Identifier: LGPL-2.1+ */
 
-#ifndef _GNU_SOURCE
-#define _GNU_SOURCE 1
-#endif
-#define __STDC_FORMAT_MACROS
+#include "config.h"
+
 #include <inttypes.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -13,7 +11,6 @@
 #include <sys/wait.h>
 #include <unistd.h>
 
-#include "config.h"
 #include "log.h"
 #include "lvm.h"
 #include "memory_utils.h"
@@ -275,10 +272,10 @@ int lvm_compare_lv_attr(const char *path, int pos, const char expected)
 		return 0;
 
 	len = strlen(output);
-	while (start < len && output[start] == ' ')
+	while ((size_t)start < len && output[start] == ' ')
 		start++;
 
-	if (start + pos < len && output[start + pos] == expected)
+	if ((size_t)(start + pos) < len && output[start + pos] == expected)
 		return 1;
 
 	return 0;
