@@ -1,8 +1,7 @@
 /* SPDX-License-Identifier: LGPL-2.1+ */
 
-#ifndef _GNU_SOURCE
-#define _GNU_SOURCE 1
-#endif
+#include "config.h"
+
 #include <libgen.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -10,10 +9,9 @@
 #include <sys/types.h>
 #include <unistd.h>
 
-#include <lxc/lxccontainer.h>
+#include "lxc.h"
 
 #include "arguments.h"
-#include "config.h"
 #include "log.h"
 
 lxc_log_define(lxc_cgroup, lxc);
